import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.net.URISyntaxException;

import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;


public class Benchmark
{
	private final class ProcessPrinter implements Runnable
	{

		private final Process process;


		public ProcessPrinter(final Process process)
		{
			this.process = process;
		}


		@Override
		public void run()
		{
			try
			{
				printOutput(process.getInputStream());
			}
			catch (IOException e)
			{
				System.err.println("Error while printing output.");
			}

		}
	}

	@Option(name = "-binPath", required = true, usage = "The path to the subversion binary directory")
	private String svnBinPath;

	@Option(name = "-repoPath", required = true, usage = "The path of the repository which will be created. All parent directories must exist.")
	private String repoPath;

	@Option(name = "-wcPath", required = true, usage = "The path of the working copy which will be created. All parent directories must exist.")
	private String wcPath;

	@Option(name = "-port", required = false, usage = "The port on which the server should be started (default is localhost:4444).")
	private int port = 4444;

	@Option(name = "-logFile", required = true, usage = "The path and name of the logFile which will be created. All parent directories must exist.")
	private String logFile;

	@Option(name = "-runs", required = false, usage = "The number of runs for each operations (default 30).")
	private int runs = 30;

	@Option(name = "-deleteRepo", required = false, usage = "Whether the created repository should be deleted after the test (default true).")
	private boolean deleteRepo = true;

	@Option(name = "-deleteWc", required = false, usage = "Whether the created working copy should be deleted after the test (default true).")
	private boolean deleteWc = true;


	public Benchmark(String[] args)
	{
		parseArguments(args);
	}


	public void execute()
	{
		init();

		Process serverProcess = null;
		try
		{
			createRepository();
			serverProcess = startServer();
			checkoutWC();
			startTest(runs);
		}
		catch (Exception e)
		{
			System.err.println("Aborted: " + e.getMessage());
		}
		finally
		{
			cleanup(repoPath, wcPath, deleteRepo, deleteWc, serverProcess);
		}
	}


	protected void init()
	{
		checkExists(svnBinPath,
					String.format("The specified subversion bin path '%s' does not exist.", svnBinPath));

		checkNotExists(repoPath,
					String.format("The specified repository path '%s' exists already.", repoPath));

		checkExists(new File(repoPath).getParent(), String.format(
					"The specified parent directory path '%s' does not exist.",
					new File(repoPath).getParent()));

		checkNotExists(wcPath,
					String.format("The specified working copy path '%s' exists already.", wcPath));

		checkExists(new File(wcPath).getParent(), String.format(
					"The specified parent directory path '%s' does not exist.",
					new File(wcPath).getParent()));

		checkExists(new File(logFile).getParent(), String.format(
					"The specified parent directory path '%s' does not exist.",
					new File(logFile).getParent()));

		if (port < 1024)
		{
			System.err.println(String.format("The specified port '%s' should be greater than 1024.",
						port));
			System.exit(1);
		}
	}


	protected void startTest(int runs) throws URISyntaxException, IOException
	{
		System.out.println(String.format("Starting test. Operations:%n" + "1. create file <file>.%n"
					+ "2. svn add <file>.%n" + "3. svn commit <file>%n" + "4. svn delete <file>%n"
					+ "5. svn commit <file>%n"));

		System.out.println("Looping " + runs + " times...");

		Timing timing = new Timing();
		String[] actions = new String[] { "create", "add", "commitadd", "delete", "commitdelete" };
		String basename = "file";

		for (String action : actions)
		{
			for (int i = 0; i < runs; i++)
			{

				addCommitAndDeleteFile("test.html", new File(wcPath + "/" + basename + i + ".html"),
							timing, action);
			}
		}

		System.out.println("Finished tests.");
		System.out.println(String.format("Results:%n"));

		PrintWriter outWriter = new PrintWriter(
					new BufferedWriter(new OutputStreamWriter(System.out)));
		writeResults(timing, actions, runs, outWriter);

		System.out.println("Writing to logFile " + logFile);
		PrintWriter fileWriter = new PrintWriter(new BufferedWriter(new FileWriter(logFile)));
		try
		{
			writeResults(timing, actions, runs, fileWriter);
		}
		finally
		{
			fileWriter.close();
		}
	}


	private void addCommitAndDeleteFile(String tstFilename, File targetFile, Timing timing,
				String action) throws IOException
	{
		timing.start();
		if ("create".equals(action))
		{
			copyFile(tstFilename, targetFile);
		}
		else if ("add".equals(action))
		{
			addFile(targetFile);
		}
		else if ("commitadd".equals(action))
		{
			commitFile(targetFile, "Commit add file");
		}
		else if ("delete".equals(action))
		{
			deleteFile(targetFile, true);
		}
		else if ("commitdelete".equals(action))
		{
			commitFile(targetFile, "Commit delete file");
		}
		timing.stop(action);
	}


	private void deleteFile(File targetFile, boolean b) throws IOException
	{
		startProcess(false, String.format("\"%s/svn\"", svnBinPath), "remove",
					targetFile.getAbsolutePath());
	}


	private void commitFile(File file, String string) throws IOException
	{
		startProcess(false, String.format("\"%s/svn\"", svnBinPath), "commit", "-m \"commitadd\"",
					file.getAbsolutePath());
	}


	private void addFile(File targetFile) throws IOException
	{
		startProcess(false, String.format("\"%s/svn\"", svnBinPath), "add",
					targetFile.getAbsolutePath());
	}


	private void copyFile(String testFilename, File dst) throws IOException
	{
		assert (!dst.exists());

		InputStream in = getClass().getClassLoader().getResourceAsStream(testFilename);
		OutputStream out = new FileOutputStream(dst.getAbsolutePath());

		try
		{
			byte[] buf = new byte[1024];
			int len;
			while ((len = in.read(buf)) > 0)
			{
				out.write(buf, 0, len);
			}
		}
		finally
		{
			try
			{
				in.close();
			}
			catch (IOException e)
			{

			}
			try
			{
				out.close();
			}
			catch (IOException e)
			{

			}
		}
	}


	protected void checkoutWC() throws IOException
	{
		startProcess(false, String.format("\"%s/svn\"", svnBinPath), "co",
					String.format("svn://localhost:%s/", port), wcPath);

		System.out.println("Checked out working copy at " + wcPath);
	}


	protected Process startServer() throws IOException
	{
		Process process = startProcess(true, String.format("\"%s/svnserve\"", svnBinPath), "-d",
					"--listen-host=localhost", String.format("--listen-port=%s", port),
					String.format("--root=%s", repoPath));

		System.out.println("Started server at localhost:" + port);
		return process;
	}


	protected void createRepository() throws IOException
	{
		startProcess(false, String.format("\"%s/svnadmin\"", svnBinPath), "create", repoPath);

		/* set anonymous write rights */
		File iniFile = new File(repoPath + "/conf/svnserve.conf");

		BufferedWriter writer = null;
		try
		{
			writer = new BufferedWriter(new FileWriter(iniFile));
			writer.write("[General]");
			writer.newLine();
			writer.write("anon-access = write");

			System.out.println("Created repository at " + repoPath);
		}
		finally
		{
			if (writer != null)
			{
				writer.close();
			}
		}

	}


	private void checkExists(String path, String errorMessage)
	{
		checkFile(path, true, errorMessage);
	}


	private void checkNotExists(String path, String errorMessage)
	{
		checkFile(path, false, errorMessage);
	}


	private void checkFile(String path, boolean mustExist, String errorMessage)
	{
		if (!checkFile(new File(path), mustExist))
		{
			System.err.println(String.format(errorMessage, new File(path).getParent()));
			System.exit(1);
		}
	}


	private boolean checkFile(File file, boolean mustExist)
	{
		return !(mustExist ^ file.exists());
	}


	private boolean deleteDirectory(File path)
	{
		if (path.exists())
		{
			File[] files = path.listFiles();
			for (int i = 0; i < files.length; i++)
			{
				if (files[i].isDirectory())
				{
					deleteDirectory(files[i]);
				}
				else
				{
					files[i].delete();
				}
			}
		}
		return (path.delete());
	}


	private Process startProcess(boolean fork, String... commands) throws IOException
	{
		ProcessBuilder cmd = new ProcessBuilder(commands);
		cmd.redirectErrorStream(true);
		Process process = cmd.start();
		if (fork)
		{
			new Thread(new ProcessPrinter(process)).start();
		}
		else
		{
			printOutput(process.getInputStream());
		}

		return process;
	}


	private void printOutput(InputStream is) throws IOException
	{
		System.out.println(convertStreamToString(is));
	}


	private String convertStreamToString(InputStream is) throws IOException
	{
		/*
		 * To convert the InputStream to String we use the Reader.read(char[] buffer)
		 * method. We iterate until the Reader return -1 which means there's no more data to
		 * read. We use the StringWriter class to produce the string.
		 */
		if (is != null)
		{
			Writer writer = new StringWriter();
			char[] buffer = new char[1024];
			try
			{
				Reader reader = new BufferedReader(new InputStreamReader(is, "UTF-8"));
				int n;
				while ((n = reader.read(buffer)) != -1)
				{
					writer.write(buffer, 0, n);
				}
			}
			finally
			{
				is.close();
			}
			return writer.toString();
		}
		else
		{
			return "";
		}
	}


	private void cleanup(String repoPath, String wcPath, boolean deleteRepo, boolean deleteWc,
				Process serverProcess)
	{
		System.out.println(String.format("Cleaning up...%n"));

		if (serverProcess != null)
		{
			serverProcess.destroy();
			System.out.println("Server shutdown.");
		}
		if (deleteRepo)
		{
			if (!deleteDirectory(new File(repoPath)))
			{
				if (new File(repoPath).exists())
				{
					System.err.println(String.format(
								"The repository created at '%s' could not be deleted.", repoPath));
				}
			}
			else
			{
				System.out.println("Repository deleted.");
			}
		}
		if (deleteWc)
		{
			if (!deleteDirectory(new File(wcPath)))
			{
				if (new File(wcPath).exists())
				{
					System.err.println(String.format(
								"The working copy created at '%s' could not be deleted.", wcPath));
				}
			}
			else
			{
				System.out.println("Working copy deleted.");
			}
		}
	}


	private void writeResults(Timing timing, String[] actions, int runs, PrintWriter writer)
				throws IOException
	{
		writer.println("+++++++++++++++++++++ Results ++++++++++++++++++++++++++");
		for (String action : actions)
		{
			writer.println(timing.get(action, runs));
		}
		writer.println("++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
		writer.flush();
	}


	private void parseArguments(String[] args)
	{
		CmdLineParser parser = new CmdLineParser(this);
		parser.setUsageWidth(80);

		try
		{
			// parse the arguments.
			parser.parseArgument(args);

		}
		catch (CmdLineException e)
		{
			System.err.println(e.getMessage());
			System.err.println("java Benchmark [options...] arguments...");

			parser.printUsage(System.err);
			System.err.println();

			System.exit(1);
		}
	}


	public static void main(String[] args)
	{
		new Benchmark(args).execute();
	}
}
